# Copyright (c) 2019 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Cinder Volume driver for Fujitsu ETERNUS DX S3 or later series."""

from hashlib import md5
import os

from oslo_log import log as logging
import paramiko

from cinder import exception
from cinder.i18n import _
from cinder import ssh_utils
from cinder.volume.drivers.fujitsu.eternus_dx import constants as CONSTANTS

LOG = logging.getLogger(__name__)


class FJPKey(paramiko.PKey):
    def get_fingerprint(self):
        """Return an MD5 fingerprint of the public part of this key.

        Declare that the use of MD5 encryption is not for security purposes.
        This declaration is to overcome connection to servers with FIPS
        security standards.
        """
        return md5(self.asbytes(), usedforsecurity=False).digest()


class FJDXSSHPool(ssh_utils.SSHPool):
    def __init__(self, *args, disabled_algorithms=None, **kwargs):
        super(FJDXSSHPool, self).__init__(*args, **kwargs)
        self.disabled_algorithms = disabled_algorithms

    def create(self):
        try:
            ssh = paramiko.SSHClient()
            if ',' in self.hosts_key_file:
                files = self.hosts_key_file.split(',')
                for f in files:
                    ssh.load_host_keys(f)
            else:
                ssh.load_host_keys(self.hosts_key_file)
            # If strict_ssh_host_key_policy is set we want to reject, by
            # default if there is not entry in the known_hosts file.
            # Otherwise we use AutoAddPolicy which accepts on the first
            # Connect but fails if the keys change.  load_host_keys can
            # handle hashed known_host entries.
            if self.strict_ssh_host_key_policy:
                ssh.set_missing_host_key_policy(paramiko.RejectPolicy())
            else:
                ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

            if self.password:
                ssh.connect(self.ip,
                            port=self.port,
                            username=self.login,
                            password=self.password,
                            timeout=self.conn_timeout)
            elif self.privatekey:
                paramiko.PKey.get_fingerprint = FJPKey.get_fingerprint
                pkfile = os.path.expanduser(self.privatekey)
                try:
                    privatekey = paramiko.ECDSAKey.from_private_key_file(
                        pkfile)
                except paramiko.ssh_exception.SSHException:
                    LOG.warning('The private key is not ECDSA. '
                                'Trying RSA.')
                    privatekey = paramiko.RSAKey.from_private_key_file(pkfile)
                ssh.connect(self.ip,
                            port=self.port,
                            username=self.login,
                            pkey=privatekey,
                            timeout=self.conn_timeout,
                            disabled_algorithms=dict(
                                pubkeys=self.disabled_algorithms))
            else:
                msg = _("Specify a password or private_key")
                raise exception.CinderException(msg)

            if self.conn_timeout:
                transport = ssh.get_transport()
                transport.set_keepalive(self.conn_timeout)
            return ssh
        except Exception as e:
            msg = _("Error connecting via ssh: %s") % str(e)
            LOG.error(msg)
            raise paramiko.SSHException(msg)


class FJDXCLI(object):
    """ETERNUS CLI Code."""

    def __init__(self, user, storage_ip, port=22, password=None, keyfile=None,
                 disabled_algorithms=None):
        """Constructor."""
        if disabled_algorithms is None:
            disabled_algorithms = []
        self.user = user
        self.storage_ip = storage_ip
        self.port = port
        if password and keyfile:
            raise Exception(_('can not specify both password and keyfile'))

        self.use_ipv6 = False
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True

        if password:
            self.ssh_pool = FJDXSSHPool(storage_ip, port, None, user,
                                        password=password, max_size=2)

        if keyfile:
            self.ssh_pool = FJDXSSHPool(storage_ip, port, None, user,
                                        disabled_algorithms=
                                        disabled_algorithms,
                                        privatekey=keyfile, max_size=2)

        self.ce_support = False
        self.CMD_dic = {
            'check_user_role': self._check_user_role,
            'expand_volume': self._expand_volume,
            'format_volume': self._format_volume,
            'format_tpv': self._format_tpv,
            'set_volume_qos': self._set_volume_qos,
            'show_volume_progress': self._show_volume_progress,
            'show_tpv_progress': self._show_tpv_progress,
            'show_host_wwn_names': self._show_host_wwn_names,
            'show_host_iscsi_names': self._show_host_iscsi_names,
            'create_host_wwn_name': self._create_host_wwn_name,
            'create_host_iscsi_name': self._create_host_iscsi_name,
            'create_affinity_group': self._create_affinity_group,
            'delete_affinity_group': self._delete_affinity_group,
            'show_affinity_group': self._show_affinity_group,
            'show_iscsi_parameters': self._show_iscsi_parameters,
            'show_fc_parameters': self._show_fc_parameters,
            'set_affinity_group': self._set_affinity_group,
            'set_host_affinity': self._set_host_affinity,
            'show_pool_provision': self._show_pool_provision,
            'show_host_affinity': self._show_host_affinity,
            'discover_host_wwn_names': self._discover_wwn_names,
            'release_host_affinity': self._release_host_affinity,
            'show_tfo_groups': self._show_tfo_groups,
            'show_storage_cluster_license': self._show_storage_cluster_license,
            'show_boxid': self._show_boxid,
            'show_copy_sessions': self._show_copy_sessions,
            'forced_tfo_group_activate': (
                self._forced_tfo_group_activate),
            'set_tfog_pair': self._set_tfog_pair,
            'show_model_supplement_info': (
                self._show_model_supplement_info),
            'show_qos_bandwidth_limit': self._show_qos_bandwidth_limit,
            'set_qos_bandwidth_limit': self._set_qos_bandwidth_limit,
            'show_volume_qos': self._show_volume_qos,
            'show_enclosure_status': self._show_enclosure_status,
            'start_copy_snap_opc': self._start_copy_snap_opc,
            'stop_copy_session': self._stop_copy_session,
            'start_copy_opc': self._start_copy_opc,
            'delete_volume': self._delete_volume,
            'show_qos_mode': self._show_qos_mode
        }

    def done(self, command, **option):
        """Done ETERNUS CLI."""
        func = self.CMD_dic.get(command, self._default_func)
        return func(**option)

    def _exec_cli(self, cmd, StrictHostKeyChecking=True, **option):

        exec_cmdline = cmd + self._get_option(**option)
        stdoutdata = self._exec_cli_with_eternus(exec_cmdline)
        output = []
        message = []
        stdoutlist = stdoutdata.split('\r\n')
        output_header = ""

        for no, outline in enumerate(stdoutlist):
            if len(outline) <= 0 or outline is None:
                continue

            if not output_header.endswith(exec_cmdline):
                output_header += outline
                continue

            if 0 <= outline.find('Error'):
                raise Exception(_("Output: %(outline)s: "
                                  "Command: %(cmdline)s")
                                % {'outline': outline,
                                   'cmdline': exec_cmdline})

            if not self._is_status(outline):
                continue

            status = int(outline, 16)
            lineno = no + 1
            break
        else:
            raise Exception(_(
                "Invalid CLI output: %(exec_cmdline)s, %(stdoutlist)s")
                % {'exec_cmdline': exec_cmdline,
                   'stdoutlist': stdoutlist})

        if status == 0:
            rc = '0'
            for outline in stdoutlist[lineno:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                message.append(outline)
        else:
            code = stdoutlist[lineno]
            for outline in stdoutlist[lineno + 1:]:
                if 0 <= outline.find('CLI>'):
                    continue
                if len(outline) <= 0:
                    continue
                if outline is None:
                    continue
                output.append(outline)
            if cmd != "show cli-error-code":
                rc, message = self._create_error_message(code, output)
            else:
                rc = 'E' + code
                message = output

        return {'result': 0, 'rc': rc, 'message': message}

    def _exec_cli_with_eternus(self, exec_cmdline):
        """Execute CLI command with arguments."""
        ssh = None
        try:
            ssh = self.ssh_pool.get()
            chan = ssh.invoke_shell()
            chan.send(exec_cmdline + '\n')
            stdoutdata = ''
            while True:
                temp = chan.recv(65535)
                if isinstance(temp, bytes):
                    temp = temp.decode('utf-8')
                else:
                    temp = str(temp)
                stdoutdata += temp

                # CLI command end with 'CLI>'.
                if stdoutdata == '\r\nCLI> ':
                    continue
                if (stdoutdata[len(stdoutdata) - 5: len(stdoutdata) - 1] ==
                        'CLI>'):
                    break
        except Exception as e:
            raise Exception(_("Execute CLI "
                              "command error. Error: %s") % str(e))
        finally:
            if ssh:
                self.ssh_pool.put(ssh)
                self.ssh_pool.remove(ssh)
        return stdoutdata

    def _show_cli_error_message(self, **option):
        """Get error messages by error code."""
        output = self._exec_cli("show cli-error-code",
                                **option)
        try:
            rc = output['rc']
            if rc != '0':
                raise Exception('_show_cli_error_message fails.')
            message = output['message'][1]
            output['message'] = message.split('\t')[1]
            return output
        except Exception as e:
            raise Exception(e)

    def _create_error_message(self, code, msg):
        """Create error code and message using arguements."""
        message = None
        rc = 'E' + code
        try:
            option = {
                'error-code': code
            }
            message = self._show_cli_error_message(**option)['message']
        except Exception:
            if rc in CONSTANTS.CLIRETCODE_dic:
                message = CONSTANTS.CLIRETCODE_dic[rc]
            else:
                message = msg

        return rc, message

    @staticmethod
    def _is_status(value):
        """Check whether input value is status value or not."""
        try:
            if len(value) != 2:
                return False

            int(value, 16)
            int(value[0], 16)
            int(value[1], 16)

            return True
        except ValueError:
            return False

    @staticmethod
    def _get_option(**option):
        """Create option strings from dictionary."""
        ret = ""
        for key, value in option.items():
            ret += " -%(key)s %(value)s" % {'key': key, 'value': value}
        return ret

    def _default_func(self, **option):
        """Default function."""
        raise Exception(_("Invalid function is specified"))

    def _check_user_role(self, **option):
        """Check user role."""
        try:
            output = self._exec_cli("show users",
                                    StrictHostKeyChecking=False,
                                    **option)
            # Return error.
            rc = output['rc']
            if rc != "0":
                return output

            userlist = output.get('message')
            role = None
            for userinfo in userlist:
                username = userinfo.split('\t')[0]
                if username == self.user:
                    role = userinfo.split('\t')[1]
                    break

            output['message'] = role
        except Exception as ex:
            if 'show users' in str(ex):
                msg = ("Specified user(%s) does not have Software role"
                       % self.user)
            elif 'Error connecting' in str(ex):
                msg = (str(ex)[34:] +
                       ', Please check fujitsu_private_key_path or .xml file')
            else:
                msg = str(ex)
            output = {
                'result': 0,
                'rc': '4',
                'message': msg
            }
        return output

    def _expand_volume(self, **option):
        """Exec expand volume."""
        return self._exec_cli("expand volume", **option)

    def _format_volume(self, **option):
        """Exec format volume."""
        return self._exec_cli("format volume", **option)

    def _format_tpv(self, **option):
        """Exec format tpv."""
        thick_option = {
            'volume-name': option['volume-name'],
            'allocation': 'thick'
        }

        output = self._exec_cli("set volume-parameters", **thick_option)
        rc = output['rc']
        if rc != "0":
            return output

        return self._exec_cli("format volume", **option)

    def _set_volume_qos(self, **option):
        """Exec set volume-qos."""
        return self._exec_cli("set volume-qos", **option)

    def _show_volume_progress(self, **option):
        """Exec show vollume-progress."""
        output = self._exec_cli("show volume-progress", **option)

        # Return error
        rc = output['rc']
        if rc != "0":
            return output

        # Get remain time.
        message = output['message']
        volume_count = message[0]
        if volume_count == "00":
            # Format completion.
            output['message'] = '0'
            return output
        else:
            remain_minute = 1
            remain_time_str = message[1].split("\t")[9]

            if remain_time_str != "FFFFFFFF":
                remain_minute = int(remain_time_str, 16) + 1

            output['message'] = str(remain_minute * 60)

        return output

    def _show_tpv_progress(self, **option):
        """Exec show tpv-progress."""
        output = self._exec_cli("show thin-pro-pools", **option)
        rc = output['rc']
        if rc != "0":
            return output

        message = output['message']
        pool_count = int(message[1], 16)
        raid_count_lineno = 2 + pool_count
        raid_count = int(message[raid_count_lineno], 16)

        for num in range(raid_count):
            raid_info = message[raid_count_lineno + num + 1]
            raid_status = int(raid_info.split('\t')[4], 16)
            cnfg_lu_status_bit_qf = raid_status & 0x0100

            if cnfg_lu_status_bit_qf != 0:
                output['message'] = '600'
                break
        else:
            output['message'] = '0'

        return output

    def _show_host_wwn_names(self, **option):
        """Get host wwn names."""
        try:
            output = self._exec_cli("show host-wwn-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[2]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host wwn names error: %s"
                           % str(ex)
            }

        return output

    def _show_host_iscsi_names(self, **option):
        """Get host iscsi names."""
        try:
            output = self._exec_cli("show host-iscsi-names", **option)

            # Return error.
            rc = output['rc']

            if rc != "0":
                return output

            hostdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                hostdatalist.append({'Host Num': int(clidata[0], 16),
                                     'Host Name': clidata[5],
                                     'IP Version': clidata[6]})

            output['message'] = hostdatalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show host iscsi names error: %s"
                           % str(ex)
            }

        return output

    def _show_affinity_group(self, **option):
        """Get affinity group information."""
        try:
            pre_count = 0
            datalist = []
            while True:
                output = self._exec_cli("show affinity-group", **option)

                rc = output['rc']

                if rc != "0":
                    return output

                clidatalist = output.get('message')

                # The specified lun group is not found
                if clidatalist[0] == '0000':
                    break
                # Get specified ag information.
                # Judge if there are more than 512 luns in the ag
                ag_info = clidatalist[1]
                ag_info_list = ag_info.split('\t')
                if len(ag_info_list) >= 4:
                    next = ag_info_list[3]
                else:
                    next = 'FFFF'

                # Get all luns in the specified ag until no subsequent ags.
                for clidataline in clidatalist[3:]:
                    # Compute the real number in the ag.
                    clidata = clidataline.split('\t')
                    num_in_ag = int(clidata[0], 16)
                    lun_num = num_in_ag + pre_count * 512
                    datalist.append({'Lun Num': lun_num,
                                     'Lun Name': clidata[2]})
                pre_count += 1
                if next == 'FFFF':
                    # No Subsequent ags connected.
                    break

                next = int(next, 16)
                option = {
                    'ag-number': next
                }

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show affinity group error: %s"
                           % str(ex)
            }

        return output

    def _show_pool_provision(self, **option):
        """Get TPP provision capacity information."""
        try:
            output = self._exec_cli("show volumes", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')

            data = 0
            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                if clidata[0] == 'FFFF':
                    break
                data += int(clidata[7], 16)
            # MB unit.
            provision = data / 2048

            output['message'] = provision
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show pool provision capacity error: %s"
                           % str(ex)
            }

        return output

    def _show_iscsi_parameters(self, **option):
        """Get iscsi port information."""

        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show iscsi-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            ipv4_data_list = []
            ipv6_data_list = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                if self.use_ipv6:
                    if clidata[31] != '::':
                        ipv6_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': '[' + clidata[31] +
                                                          ']',
                                               'port str': port_str})
                else:
                    if clidata[10] != '0.0.0.0':
                        ipv4_data_list.append({'port num': tmpport,
                                               'port iqn': clidata[4],
                                               'port ip': clidata[10],
                                               'port str': port_str})

            if self.use_ipv6:
                output['message'] = ipv6_data_list
            else:
                output['message'] = ipv4_data_list
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show iscsi-parameters error: %s"
                           % str(ex)
            }

        return output

    def _show_fc_parameters(self, **option):
        """Get fc port information."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show fc-parameters", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                tmpport = clidata[0] + clidata[1]
                port_str = self._get_port(clidata[0], clidata[1])
                datalist.append({'port num': tmpport,
                                 'port wwpn': clidata[18],
                                 'port str': port_str})

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show fc-parameters error: %s"
                           % str(ex)
            }

        return output

    def _get_port(self, ca_module_id, port_number):
        if self.ce_support:
            return self._get_port_str_v2(ca_module_id, port_number)
        else:
            return self._get_port_str(ca_module_id, port_number)

    @staticmethod
    def _get_port_str(ca_module_id, port_number):
        cm = int(ca_module_id[1], 16) % 8
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) // 4) // 2) * 2 + (
            int(port_number[1], 16) % 2)
        temp_str = str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    @staticmethod
    def _get_port_str_v2(ca_module_id, port_number):
        # Port_str format: CE + CM + CA + PORT
        # CE: 0,1,2,3,4,5,6,7,8,9,10,11
        # CM: 0,1
        # CA: 0,1,2,3
        # PORT: 0,1,2,3
        base = 4
        ce = ((int(ca_module_id[1], 16) % 8) // 2) + (
            ((int(ca_module_id[0], 16) - base) // 4) * 4)
        cm = int(ca_module_id[1], 16) % 2
        ca = int(ca_module_id[0], 16) % 4
        port = ((int(ca_module_id[1], 16) // 4) // 2) * 2 + (
            int(port_number[1], 16) % 2)
        temp_str = str(hex(ce)) + str(hex(cm)) + str(hex(ca)) + str(hex(port))
        return temp_str.replace('0x', '').upper()

    def _create_host_wwn_name(self, **option):
        """Create host wwn name."""
        return self._exec_cli("create host-wwn-name", **option)

    def _create_host_iscsi_name(self, **option):
        """Create host iscsi name."""
        return self._exec_cli("create host-iscsi-name", **option)

    def _create_affinity_group(self, **option):
        """Create affinity group."""
        return self._exec_cli("create affinity-group", **option)

    def _delete_affinity_group(self, **option):
        """Delete affinity group."""
        return self._exec_cli("delete affinity-group", **option)

    def _set_host_affinity(self, **option):
        """Set host affinity."""
        return self._exec_cli("set host-affinity", **option)

    def _set_affinity_group(self, **option):
        """Set affinity group."""
        return self._exec_cli("set affinity-group", **option)

    def _show_host_affinity(self, **option):
        """Get host affinity group information."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("show host-affinity", **option)

            rc = output['rc']

            if rc != "0":
                return output

            # [{Port:xxx,Host Num:xxx,Host Name:xxx,AG Num:xxx,AG Name:xxx}].
            datalist = []
            clidatalist = output.get('message')

            port_str = None
            for line in clidatalist:
                infos = line.split()
                if len(infos) == 2:
                    port_str = self._get_port(
                        infos[0][-4:-2], infos[0][-2:])
                if len(infos) == 7 or len(infos) == 6:
                    host_ag = {
                        'Port': port_str,
                        'Host Num': int(infos[0], 16),
                        'Host Name': infos[1],
                        'AG Num': int(infos[2], 16),
                        'AG Name': infos[3]
                    }
                    datalist.append(host_ag)

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % str(ex)
            }

        return output

    def _discover_wwn_names(self, **option):
        """Get FC zone info on ETERNUS."""
        if 'ce_support' in option:
            if option['ce_support']:
                self.ce_support = True
            option.pop('ce_support')

        try:
            output = self._exec_cli("discover host-wwn-names", **option)

            rc = output['rc']

            if rc != "0":
                return output

            datalist = {}  # {Host WWPN: [port1, port2]}.
            clidatalist = output.get('message')

            port_str = None
            for idx, line in enumerate(clidatalist):
                infos = line.split()
                if len(infos) == 3:
                    port_str = self._get_port(infos[0], infos[1])
                    port_idx = idx
                    continue
                if port_str and idx != port_idx + 1:
                    wwpn = infos[0]
                    temp_list = datalist.get(wwpn, [])
                    temp_list.append(port_str)
                    datalist[wwpn] = temp_list

            output['message'] = datalist
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "discover host-wwn-names: %s"
                           % str(ex)
            }

        return output

    def _release_host_affinity(self, **option):
        """Release host affinity."""
        return self._exec_cli("release host-affinity", **option)

    def _show_tfo_groups(self, **option):
        """Show tfo groups."""
        try:
            local_ce_support = option['local_ce_support']
            remote_ce_support = option['remote_ce_support']
            option.pop('local_ce_support')
            option.pop('remote_ce_support')
            output = self._exec_cli("show tfo-groups", **option)

            rc = output['rc']

            if rc != "0":
                return output

            tfog_list = []
            clidatalist = output.get('message')
            datalist = clidatalist[1:]

            for index, item in enumerate(datalist):
                if len(item) == 4:
                    tfog_port_list = []
                    tmp = []
                    tfog_dict = {}
                    # Get TFO group information.
                    tfog_info = datalist[index - 1]
                    tfog_line = tfog_info.strip().split('\t')
                    tfog_dict['no'] = tfog_line[0]
                    tfog_dict['name'] = tfog_line[1]
                    tfog_dict['type'] = tfog_line[2]
                    tfog_dict['status'] = tfog_line[4]
                    tfog_dict['phase'] = tfog_line[5]
                    tfog_dict['condition'] = tfog_line[6]
                    tfog_dict['pair_boxid'] = tfog_line[8]

                    # Get each port-pair information.
                    tmp = [
                        datalist[index + i + 1] for i in range(int(item))
                    ]
                    for sub_tmp in tmp:
                        tfog_port_dict = {}
                        tfog_ports = sub_tmp.strip().split('\t')
                        if local_ce_support:
                            tfog_port_dict['own_port'] = tfog_ports[0]
                        else:
                            tfog_port_dict['own_port'] = tfog_ports[0][-3:]
                        # Local storage only has one remote storage.
                        if remote_ce_support:
                            tfog_port_dict['pair_port'] = tfog_ports[1]
                        else:
                            tfog_port_dict['pair_port'] = tfog_ports[1][-3:]
                        if tfog_ports[2] == '02':
                            tfog_port_dict['port_prtcl'] = 'iSCSI'
                        else:
                            tfog_port_dict['port_prtcl'] = 'FC'
                        tfog_port_list.append(tfog_port_dict)

                    tfog_dict['tfog_port_list'] = tfog_port_list
                    tfog_list.append(tfog_dict)

            output['message'] = tfog_list
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show tfo group error: %s"
                           % str(ex)
            }

        return output

    def _show_storage_cluster_license(self, **option):
        """Show Storage Cluster license"""
        try:
            output = self._exec_cli("show storage-cluster-license", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            data = clidatalist[0]

            if data == '01':
                license = 'Registered'
            else:
                license = 'Disable'

            output['message'] = license

        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show storage-cluster-license error: %s"
                           % str(ex)
            }

        return output

    def _show_boxid(self, **option):
        """Show boxid."""
        try:
            output = self._exec_cli("show boxid", **option)

            rc = output['rc']

            if rc != "0":
                return output

            output['message'] = output.get('message')[0]
        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show boxid error: %s"
                           % str(ex)
            }

        return output

    def _forced_tfo_group_activate(self, **option):
        """Failover or Failback"""
        return self._exec_cli("forced tfo-group-activate", **option)

    def _set_tfog_pair(self, **option):
        """set tfog pair and create copy sessions"""
        return self._exec_cli("set tfo-pair", **option)

    def _show_model_supplement_info(self, **option):
        """Show model supplement info."""
        try:
            output = self._exec_cli("show model-supplement-info", **option)

            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            infos = clidatalist[0].split()
            output['message'] = {
                'machine_info': infos[0],
                'model_info': infos[1]}

        except Exception as ex:
            output = {
                'result': 0,
                'rc': '4',
                'message': "show model-supplement-info error: %s"
                           % str(ex)
            }

        return output

    def _show_copy_sessions(self, **option):
        """Get copy sessions."""
        try:
            output = self._exec_cli("show copy-sessions", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            cpsdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                # Get CopyType
                if clidata[2] == '01':
                    # CopyKind: OPC
                    if bin(int(clidata[3], 16) & 16) != 0:
                        # eg. 0b10010000
                        temp_type = 'Snap'
                    elif bin(int(clidata[3], 16) & 64) != 0:
                        # eg. 0b11000000
                        temp_type = 'Snap+'
                    else:
                        temp_type = 'Other'
                elif clidata[2] == '02':
                    # CopyKind: EC
                    if clidata[5] == 'FF':
                        temp_type = 'EC'
                    elif clidata[5] == '10':
                        temp_type = 'Sync_REC'
                    else:
                        temp_type = 'Other'
                else:
                    temp_type = 'Other'

                # Get Destination Session ID when session is REC.
                if clidata[1] != 'FFFF':
                    temp_dest_session_id = int(clidata[1], 16)
                else:
                    temp_dest_session_id = int(clidata[0], 16)

                # Get Role in Remote Flag when copy session is REC
                if clidata[2] == '02':
                    # CopyKind: EC
                    if clidata[5] == 'FF':
                        temp_role = 'Other'
                    else:
                        # Only REC copy has the remote flag
                        if int(clidata[4], 16) & 16:
                            temp_role = 'Destination'
                        else:
                            temp_role = 'Source'
                else:
                    temp_role = 'Other'

                # Get Phases
                if clidata[6] == '00':
                    temp_phase = 'No_Pair'
                elif clidata[6] == '01':
                    temp_phase = 'Copying'
                elif clidata[6] == '02':
                    temp_phase = 'Equivalent'
                elif clidata[6] == '03':
                    temp_phase = 'Tracking'
                elif clidata[6] == '04':
                    temp_phase = 'Tracking_Copying'
                elif clidata[6] == '06':
                    temp_phase = 'Readying'
                else:
                    temp_phase = 'Other'

                # Get CopyStatus
                if clidata[7] == '00':
                    temp_status = 'Idle'
                elif clidata[7] == '01':
                    temp_status = 'Reserve'
                elif clidata[7] == '02':
                    temp_status = 'Active'
                elif clidata[7] == '03':
                    temp_status = 'Error_Suspend'
                elif clidata[7] == '04':
                    temp_status = 'Suspend'
                elif clidata[7] == '05':
                    temp_status = 'Halt'
                else:
                    temp_status = 'Other'

                cpsdatalist.append({'Source Num': int(clidata[13], 16),
                                    'Dest Num': int(clidata[14], 16),
                                    'Role': temp_role,
                                    'Type': temp_type,
                                    'Status': temp_status,
                                    'Phase': temp_phase,
                                    'Session ID': int(clidata[0], 16),
                                    'Dest Session ID': temp_dest_session_id})

            output['message'] = cpsdatalist
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show copy sessions error: %s"
                                 % str(ex)}

        return output

    def _show_qos_bandwidth_limit(self, **option):
        """Get qos bandwidth limit."""
        try:
            output = self._exec_cli("show qos-bandwidth-limit", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            qoslist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                qoslist.append({'total_limit': int(clidata[0], 16),
                                'total_iops_sec': int(clidata[1], 16),
                                'total_bytes_sec': int(clidata[2], 16),
                                'read_limit': int(clidata[0], 16),
                                'read_iops_sec': int(clidata[3], 16),
                                'read_bytes_sec': int(clidata[4], 16),
                                'write_limit': int(clidata[0], 16),
                                'write_iops_sec': int(clidata[5], 16),
                                'write_bytes_sec': int(clidata[6], 16)})

            output['message'] = qoslist

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show qos bandwidth limit error: %s"
                                 % str(ex)}

        return output

    def _set_qos_bandwidth_limit(self, **option):
        """Set qos bandwidth limit"""
        return self._exec_cli("set qos-bandwidth-limit", **option)

    def _show_volume_qos(self, **option):
        """Get volumes with qos."""
        try:
            output = self._exec_cli("show volume-qos", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            vqosdatalist = []
            clidatalist = output.get('message')

            for clidataline in clidatalist[1:]:
                clidata = clidataline.split('\t')
                vqosdatalist.append({'total_limit': int(clidata[2], 16),
                                     'read_limit': int(clidata[3], 16),
                                     'write_limit': int(clidata[4], 16)})

            output['message'] = vqosdatalist

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show volume qos error: %s"
                                 % str(ex)}

        return output

    def _show_enclosure_status(self, **option):
        """Get the version of machine."""
        try:
            output = self._exec_cli("show enclosure-status", **option)

            # return error
            rc = output['rc']

            if rc != "0":
                return output

            clidatalist = output.get('message')
            clidata = clidatalist[0].split('\t')
            versioninfo = {'version': clidata[11]}

            output['message'] = versioninfo

        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show enclosure status error: %s"
                                 % str(ex)}

        return output

    def _start_copy_snap_opc(self, **option):
        """Exec start copy-snap-opc."""
        return self._exec_cli("start copy-snap-opc", **option)

    def _stop_copy_session(self, **option):
        """Exec stop copy-session."""
        return self._exec_cli("stop copy-session", **option)

    def _start_copy_opc(self, **option):
        """Exec start copy-opc."""
        return self._exec_cli("start copy-opc", **option)

    def _delete_volume(self, **option):
        """Exec delete volume."""
        return self._exec_cli('delete volume', **option)

    def _show_qos_mode(self, **option):
        """Get QoS mode."""
        try:
            output = self._exec_cli("show qos-mode", **option)

            rc = output['rc']
            if rc != "0":
                return output

            clidatalist = output.get('message')[0].split('\t')
            if len(clidatalist) >= 6:
                bws_mode = int(clidatalist[0], 16)
                priority_mode = int(clidatalist[2], 16)
                if bws_mode == 1 and priority_mode == 1:
                    output['message'] = 'QoS Priority Enable'
                elif bws_mode == 1 and priority_mode == 0:
                    output['message'] = 'BandWidth Limit Enable'
                else:
                    output['message'] = 'Disable'
            else:
                bws_mode = int(clidatalist[0], 16)
                if bws_mode:
                    output['message'] = 'BandWidth Limit Enable'
                else:
                    output['message'] = 'Disable'
        except Exception as ex:
            output = {'result': 0,
                      'rc': '4',
                      'message': "Show qos mode error: %s"
                                 % str(ex)}
        return output
