from cinder.volume.drivers.fujitsu.eternus_dx.osvd_monitor \
    import log
import socket
import os
import re
from threading import Timer
import time
import pymysql
import configparser
import prettytable as pt
from xml.etree.ElementTree import parse

from cinder.i18n import _
from cinder.volume.drivers.fujitsu.eternus_dx.osvd_monitor \
    import eternus_dx_monitor_cli
from cinder.volume.drivers.fujitsu.eternus_dx.osvd_monitor \
    import osvd_host_affinity_monitor as monitor
from oslo_concurrency import lockutils
from oslo_config import cfg

QoS_CFG_FILE = 'qos_check.conf'
OSVD_CONF = "osvd_conf"
PASSWORDLESS = "passwordless"
PRIVATE_KEY = ""
DEFAULT_PRIVATE_KEY = "/var/lib/cinder/eternus"
DB_USER = 'cinder_database_user'
DB_PASS = 'cinder_database_pass'

LOG = log.FinalLogger().getLogger(__name__)

cfg_dic = monitor.get_value_from_cfg_file(QoS_CFG_FILE)
db_user = cfg_dic.get(DB_USER, None)
db_pass = cfg_dic.get(DB_PASS, None)
hostname = socket.getfqdn(socket.gethostname())

FJ_QOS_KEY_list = [
    'maxBWS'
]
FJ_QOS_KEY_BYTES_list = [
    'read_bytes_sec',
    'write_bytes_sec',
    'total_bytes_sec'
]
FJ_QOS_KEY_IOPS_list = [
    'read_iops_sec',
    'write_iops_sec',
    'total_iops_sec'
]
FJ_PRIORITY_KEY_list = [
    'volume_priority'
]


class FJDXCheckQoS(monitor.FJDXMonitor):
    def __init__(self, **kwargs):
        self.config_file = kwargs.get('config_file')
        passwordless = kwargs.get('passwordless')
        private_key_path = kwargs.get('private_key_path')

        self.use_ipv6 = False
        storage_ip = self._get_drvcfg('EternusIP', filename=self.config_file)
        if storage_ip.find(':') != -1:
            self.use_ipv6 = True

        if passwordless.lower() == 'false':
            self.passwordless = False
        else:
            self.passwordless = True

        if private_key_path:
            self.private_key_path = private_key_path
        else:
            self.private_key_path = DEFAULT_PRIVATE_KEY

        if disabled_algorithms:
            self.disabled_algorithms = disabled_algorithms
        else:
            self.disabled_algorithms = None

        if ssh_port:
            self.ssh_port = ssh_port
        else:
            self.ssh_port = 22

        self.fjdxcli = None

    def _get_qos_mode(self):
        LOG.debug('_get_qos_mode, Enter method.')

        rc, errordesc, qos_mode_list = self._exec_eternus_cli('show_qos_mode')
        if rc != 0:
            msg = (_('_get_qos_mode, '
                     'show_qos_mode failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'clidata': qos_mode_list})
            LOG.error(msg)
            raise Exception(msg)

        LOG.debug('_get_qos_mode, qos mode: %(qos_mode_list)s.',
                  {'qos_mode_list': qos_mode_list})
        return qos_mode_list

    def _get_enclosure_status(self):
        LOG.debug('_get_enclosure_status, Enter method.')

        rc, errordesc, clidata = \
            self._exec_eternus_cli('show_enclosure_status')

        if rc != 0:
            msg = (_('_get_enclosure_status, '
                     'show_qos_mode failed. '
                     'Return code: %(rc)lu, '
                     'Error: %(errordesc)s, '
                     'Message: %(clidata)s.')
                   % {'rc': rc,
                      'errordesc': errordesc,
                      'clidata': clidata})
            LOG.error(msg)
            raise Exception(msg)

        firm_version = clidata['version']
        LOG.debug('_get_enclosure_status, firm version: %(version)s.',
                  {'version': firm_version})

        if firm_version < 'V11L30':
            version = 'old'
        else:
            version = 'new'

        return version


def _get_qos_list():
    # Get QoS list from database
    LOG.debug('_get_qos_list, Entermethod')

    qos_dict = {}

    db = pymysql.connect(host=hostname, user=db_user, password=db_pass,
                         database='cinder')
    cursor = db.cursor()
    sql = 'select specs_id, quality_of_service_specs.key, value ' \
          'from quality_of_service_specs where deleted=0 and ' \
          'quality_of_service_specs.key!="consumer" and ' \
          'quality_of_service_specs.key!="QoS_Specs_Name"'
    try:
        cursor.execute(sql)
        results = cursor.fetchall()
        qoss = results
    except Exception as ex:
        msg = ('_get_qos_list, get qos list failed. Error message: {ex}.',
               {'ex': ex})
        LOG.error(msg)
        raise Exception(msg)
    finally:
        db.close()

    for qos in qoss:
        if qos_dict.get(qos[0], None) is None:
            qos_dict[qos[0]] = {}
        qos_dict[qos[0]][qos[1]] = qos[2]

    return qos_dict


def _get_type_list():
    # Get type list from database.
    LOG.debug('_get_type_list, Enter method')

    type_dict = {}

    db = pymysql.connect(host=hostname, user=db_user, password=db_pass,
                         database='cinder')
    cursor = db.cursor()
    sql1 = 'select id, name, qos_specs_id from volume_types where deleted=0'
    sql2 = 'select volume_type_id, volume_type_extra_specs.key, value ' \
           'from volume_type_extra_specs where deleted=0'

    try:
        cursor.execute(sql1)
        results = cursor.fetchall()
        types = results
        cursor.execute(sql2)
        results = cursor.fetchall()
        extras = results
    except Exception:
        raise Exception
    finally:
        db.close()

    for tp in types:
        if not type_dict.get(tp[0], None):
            type_dict[tp[0]] = {}
        type_dict[tp[0]]['name'] = tp[1]
        type_dict[tp[0]]['qos_specs_id'] = tp[2]

    for extra in extras:
        type_dict[extra[0]][extra[1]] = extra[2]

    return type_dict


def _get_volume_list():
    # Get volume list from database.
    LOG.debug('_get_volume_list, Enter method.')

    volume_dict = dict()

    db = pymysql.connect(host=hostname, user=db_user, password=db_pass,
                         database='cinder')
    cursor = db.cursor()

    sql = 'select id, display_name, volume_type_id ' \
          'from volumes where deleted=0 and status!="error" ' \
          'and status!="error_deleting" and status!="deleting"' \
          'and status!="error_extending" and status!="error_managing" ' \
          'and status!="error_restoring" and status!="error_backing-up"'
    try:
        cursor.execute(sql)
        results = cursor.fetchall()
        volumes = results
    except:
        raise Exception
    finally:
        db.close()

    for volume in volumes:
        if not volume_dict.get(volume[0], None):
            volume_dict[volume[0]] = dict()
        volume_dict[volume[0]]['name'] = volume[1]
        volume_dict[volume[0]]['type'] = volume[2]

    return volume_dict


def _get_conf_para():
    LOG.debug('_get_conf_para, Enter method.')

    cf = configparser.ConfigParser()
    cf.read('/etc/cinder/cinder.conf')
    secs = cf.sections()
    cfg_dicts = dict()

    # Get enabled backend.
    enabled_backend = cf.get("DEFAULT", 'enabled_backends')
    enabled_backend = enabled_backend.replace(" ", "")
    enabled_backend = enabled_backend.split(',')
    for sec in secs:
        tmp = dict()
        items = cf.items(sec)
        if sec in enabled_backend:
            # If backend is enabled, find the backend information.
            for item in items:
                # item like this: (fujitsu_password_less, True)
                if isinstance(item, tuple) and len(item) == 2:
                    # Add item into dictionary. Like this:
                    # {'fujitsu_password_less': 'True'}
                    tmp[item[0]] = item[1]
                    if item[0] == 'volume_driver':
                        # Determine the backend driver.
                        # volume_driver = xxx
                        if 'FJDXISCSIDriver' in item[1] or \
                                'FJDXFCDriver' in item[1]:
                            tmp['is_fujitsu'] = True
                        else:
                            tmp['is_fujitsu'] = False
            cfg_dicts[sec] = tmp
        else:
            # Backend is disabled, find the volume_backend_name.
            volume_backend_name = 'None'
            is_fujitsu = False
            for item in items:
                if isinstance(item, tuple) and len(item) == 2:
                    if item[0] == 'volume_backend_name':
                        volume_backend_name = item[1]
                    if item[0] == 'volume_driver':
                        if 'FJDXISCSIDriver' in item[1] or \
                                'FJDXFCDriver' in item[1]:
                            is_fujitsu = True
            cfg_dicts[sec] = {'volume_backend_name': volume_backend_name,
                              'is_fujitsu': is_fujitsu}

    return enabled_backend, cfg_dicts


def type_check(qos_list, type_list, cfg_dict, qos_mode_dict,
               enabled_backend):
    LOG.debug('type_ckeck, Enter method.')
    type_states = {}
    for types, extra in type_list.items():
        type_states[types] = {}
        if 'volume_backend_name' not in extra:
            # Type does not bind any backend.
            type_states[types]['status'] = '-'
            type_states[types]['reason'] = \
                'Volume type does not bind any backend.'
            continue

        backend_name = extra['volume_backend_name']
        cfg_section = None
        is_fujitsu = False
        for section, option in cfg_dict.items():
            # Check if the volume type bind a effective backend
            if backend_name == option.get('volume_backend_name', None):
                cfg_section = section
                if option.get('is_fujitsu', None):
                    is_fujitsu = option['is_fujitsu']
                break
        if not cfg_section:
            # Volume type bind an error backend.
            type_states[types]['status'] = ':('
            type_states[types]['reason'] = \
                'Volume type does not bind a effective backend.'
            continue
        elif not is_fujitsu:
            # Volume type does not bind a fujitsu backend
            type_states[types]['status'] = '-'
            type_states[types]['reason'] = \
                'Volume type binds None Fujitsu backend.'
        else:
            if cfg_section not in enabled_backend:
                # The volume backend is not start.
                type_states[types]['status'] = ':('
                type_states[types]['reason'] = 'Volume backend is disabled.'
                continue
            else:
                # The volume backend is start.
                if not extra['qos_specs_id']:
                    # The volume type does not associate to any qos.
                    type_states[types]['status'] = ':)'
                    type_states[types]['reason'] = '-'
                    continue
                else:
                    xml_path = cfg_dict[cfg_section][
                        'cinder_eternus_config_file']
                    qos_mode = qos_mode_dict[xml_path]
                    qos_id = extra['qos_specs_id']
                    qos_spec = qos_list[qos_id]
                    if 'BandWidth Limit Enable' in qos_mode:
                        for qos_key in qos_spec.keys():
                            # Storage QoS mode is bws, but the information
                            # of qos that type associated is qos priority.
                            if qos_key in FJ_PRIORITY_KEY_list:
                                type_states[types]['status'] = ':('
                                reason = (_('Volume type cannot be used '
                                            'because the QoS information '
                                            'of the volume type and '
                                            'storage QoS mode is inconsistent.'
                                            '\n(Storage: BandWidth Limit,'
                                            'now: QoS priority.)'))
                                type_states[types]['reason'] = reason

                            if qos_key in FJ_QOS_KEY_IOPS_list or \
                                    qos_key in FJ_QOS_KEY_BYTES_list:
                                if qos_mode.startswith('old'):
                                    type_states[types]['status'] = ':('
                                    reason = (_('Volume type cannot be used '
                                                'because the QoS information '
                                                'of the volume type and '
                                                'storage QoS mode is '
                                                'inconsistent.'
                                                '\n(Storage: maxBWS support,'
                                                'now: iops use.)'))
                                    type_states[types]['reason'] = reason

                            if qos_key in FJ_QOS_KEY_list:
                                if qos_mode.startswith('new'):
                                    type_states[types]['status'] = ':('
                                    reason = (_('Volume type cannot be used '
                                                'because the QoS information '
                                                'of the volume type and '
                                                'storage QoS mode is '
                                                'inconsistent.'
                                                '\n(Storage: iops support,'
                                                'now: maxBWS use.)'))
                                    type_states[types]['reason'] = reason

                        if not type_states[types].get('status', None):
                            type_states[types]['status'] = ':)'
                            reason = '-'
                            type_states[types]['reason'] = reason
                    elif 'QoS Priority Enable' in qos_mode:
                        for qos_key in qos_spec.keys():
                            # Storage QoS mode is qos priority, but the
                            # information of qos that type associated is bws.
                            if qos_key in FJ_QOS_KEY_list or \
                                    qos_key in FJ_QOS_KEY_IOPS_list or \
                                    qos_key in FJ_QOS_KEY_BYTES_list:
                                type_states[types]['status'] = ':('
                                reason = (_('Volume type cannot be used '
                                            'because the QoS information '
                                            'of the volume type and '
                                            'storage QoS mode is inconsistent.'
                                            '\n(Storage: QoS Priority,'
                                            'now: BandWidth Limit.'))
                                type_states[types]['reason'] = reason
                        if not type_states[types].get('status', None):
                            type_states[types]['status'] = ':)'
                            reason = '-'
                            type_states[types]['reason'] = reason
                    elif 'Disable' in qos_mode:
                        # Storage does not set any QoS mode.
                        for qos_key in qos_spec.keys():
                            if qos_key in FJ_QOS_KEY_list or \
                                    qos_key in FJ_QOS_KEY_IOPS_list or \
                                    qos_key in FJ_QOS_KEY_BYTES_list or \
                                    qos_key in FJ_PRIORITY_KEY_list:
                                type_states[types]['status'] = ':('
                                reason = (_('Volume type cannot be used '
                                            'because the QoS information '
                                            'of the volume type and '
                                            'storage QoS mode is inconsistent.'
                                            '\n(Storage: Disable,'
                                            'now: BandWidth Limit or '
                                            'QoS priority.'))
                                type_states[types]['reason'] = reason
                        if not type_states[types].get('status', None):
                            type_states[types]['status'] = ':)'
                            reason = '-'
                            type_states[types]['reason'] = reason
                    continue
    return type_states


def volume_check(volume_dict, type_states, type_list, qos_list):
    LOG.debug('volume_check, Enter method.')
    volume_need_retype = list()
    for volume, params in volume_dict.items():
        volumes_dict = dict()
        volume_type_id = params.get('type', None)
        if volume_type_id:
            for tp, chk in type_states.items():
                if tp != volume_type_id:
                    continue
                else:
                    type_status = chk['status']
                    if type_status == ':(':
                        volumes_dict['id'] = volume
                        volumes_dict['type'] = volume_type_id
                        volumes_dict['name'] = params.get('name')
                        qos_id = type_list[volume_type_id]['qos_specs_id']
                        volume_qos = qos_list.get(qos_id, None)
                        volumes_dict['qos'] = volume_qos
                        volume_need_retype.append(volumes_dict)
    return volume_need_retype


if __name__ == '__main__':
    qos_list = _get_qos_list()
    type_list = _get_type_list()
    enabled_backend, cfg_dict = _get_conf_para()

    qos_mode_dict = {}
    backend_info = {}

    for key, values in cfg_dict.items():
        if isinstance(values, dict):
            if values['is_fujitsu'] and key in enabled_backend:
                backend_name = values.get('volume_backend_name', None)
                xml_file = values.get('cinder_eternus_config_file', None)
                passwordless = values.get('fujitsu_passwordless', 'True')
                private_key_path = values.get('fujitsu_private_key_path',
                                              PRIVATE_KEY)
                ssh_port = values.get('fujitsu_ssh_port', 22)
                disabled_algorithms = values.get('fujitsu_disabled_algorithms',
                                                 None)
                backend_info = {
                    'backend_name': backend_name,
                    'config_file': xml_file,
                    'passwordless': passwordless,
                    'private_key_path': private_key_path,
                    'ssh_port': ssh_port,
                    'disabled_algorithms': disabled_algorithms
                }
                backend_monitor = FJDXCheckQoS(**backend_info)
                qos_mode = backend_monitor._get_qos_mode()
                firm_version = backend_monitor._get_enclosure_status()
                version_mode = firm_version + '-' + qos_mode
                qos_mode_dict[xml_file] = version_mode

    type_ok = type_check(qos_list, type_list, cfg_dict, qos_mode_dict,
                         enabled_backend)

    tb_retype = pt.PrettyTable()
    tb_retype.field_names = ['type ID', 'type name', 'associated QoS',
                             'binded backend', 'safe to use',
                             'reason for unsafe']
    for types, check in type_ok.items():
        qos_id = type_list[types]['qos_specs_id'] \
            if type_list[types]['qos_specs_id'] else '-'
        backend_name = type_list[types]['volume_backend_name'] \
            if 'volume_backend_name' in type_list[types] else None
        tb_retype.add_row([types, type_list[types]['name'], qos_id,
                           backend_name, check['status'], check['reason']])

    print(tb_retype)

    volume_list = _get_volume_list()
    need_retype = volume_check(volume_list, type_ok, type_list, qos_list)
    tb_volume = pt.PrettyTable()
    tb_volume.field_names = ['volume ID', 'volume name', 'volume type',
                             'volume qos spec']
    for volume in need_retype:
        type_id = volume.get('type', None)
        type_name = type_list[type_id]['name'] if type_id else '-'
        tb_volume.add_row([volume['id'], volume['name'], type_name,
                           volume['qos']])
    print(tb_volume)
